/**
  *  \file client/dialogs/simfunction.cc
  *  \brief Simulator Ship Function Editor
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "client/dialogs/simfunction.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/optiongrid.h"
#include "game/sim.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/strtab.h"
#include "ui/widgets/quit.h"

namespace {
    const int NUM_FUNCTIONS = 10;

    class WSimFunctionEditor : public UIWindow {
        GSimObject& obj;
        ObjectHolder holder;
        int new_flags;
        WOptionGrid* pgrid;

        void toggle(GSimFunction sf);
        void init(bool ship);
     public:
        WSimFunctionEditor(const char* ttl, bool ship, GSimObject& obj);
        void onClick(int i);
        void render();
        const char* getFunctionType(GSimFunction sf) const;
        int getNewFlags() const
            { return new_flags; }
    };
}

static const char KEYS[NUM_FUNCTIONS] = { 'I', 'F', 'C', 'K', 'B', 'T', 'E', 'Q', 'G', 'Y' };

static const char* const LABELS[NUM_FUNCTIONS] = {
    N_("Planet Immunity"),
    N_("Full Weaponry"),
    N_("Commander"),
    N_("3\xC3\x97 Beam Kill"),
    N_("2\xC3\x97 Beam Charge"),
    N_("2\xC3\x97 Torp Charge"),
    N_("Elusive"),
    N_("Squadron"),
    N_("Shield Generator"),
    N_("Cloaked Fighter Bays"),
};

static const char* const VALUES[] = {
    N_("no"),
    N_("yes"),
    N_("default (=no)"),
    N_("default (=yes)"),
};

WSimFunctionEditor::WSimFunctionEditor(const char* ttl, bool ship, GSimObject& obj)
    : UIWindow(ttl, BLUE_WINDOW, COLOR_GRAY),
      obj(obj),
      new_flags(obj.getFlags()),
      pgrid(0)
{
    init(ship);
}

void
WSimFunctionEditor::init(bool ship)
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& buttons = holder.add(new UIGroup(UIHBoxLayout::instance5));

    WOptionGrid& grid = holder.add(new WOptionGrid(getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(LABELS, countof(LABELS))),
                                                   getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(VALUES, countof(VALUES)))));

    // Populate it
    for (int i = 0; i < NUM_FUNCTIONS; ++i) {
        // Show item if it is a ship (all items)
        // or one of the items supported for planets.
        if (ship || (i == sf_TripleBeamKill || i == sf_DoubleBeamCharge || i == sf_DoubleTorpCharge)) {
            grid.addItem(i, KEYS[i], _(LABELS[i]));
        }
    }
    grid.sig_click.add(this, &WSimFunctionEditor::onClick);

    buttons.add(holder.add(new UIButton(cm_Enter,  _("OK"), SDLK_RETURN)).addStop());
    buttons.add(holder.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    buttons.add(holder.add(new UISpacer()));
    buttons.add(holder.add(new UIButton(_("H"), 'h')));

    add(grid);
    add(buttons);
    add(holder.add(new WHelpWidget("pcc2:simfunctions")));
    add(holder.add(new UIQuit(cm_Escape)));
    pack();

    pgrid = &grid;
    render();
}

void
WSimFunctionEditor::toggle(GSimFunction sf)
{
    const int32_t value = GSimObject::simf_to_value_bit[sf];
    const int32_t set = GSimObject::simf_to_set_bit[sf];
    const int32_t flags = new_flags;
    if ((flags & set) == 0) {
        // default => Yes
        new_flags |= (value | set);
    } else if ((flags & value) != 0) {
        // Yes => No
        new_flags &= ~value;
    } else {
        // No => Default
        new_flags &= ~set;
    }

    render();
}

void
WSimFunctionEditor::onClick(int i)
{
    toggle(GSimFunction(i));
}

void
WSimFunctionEditor::render()
{
    if (pgrid) {
        for (int i = 0; i < NUM_FUNCTIONS; ++i) {
            pgrid->find(i).setValue(getFunctionType(GSimFunction(i)));
        }
    }
}

const char*
WSimFunctionEditor::getFunctionType(GSimFunction sf) const
{
    const int32_t value = GSimObject::simf_to_value_bit[sf];
    const int32_t set = GSimObject::simf_to_set_bit[sf];
    const int32_t flags = new_flags;

    int index = 0;
    if ((flags & set) != 0) {
        if ((flags & value) != 0) {
            index = 1;
        } else {
            index = 0;
        }
    } else {
        if (obj.hasImpliedFunction(sf)) {
            index = 3;
        } else {
            index = 2;
        }
    }
    return _(VALUES[index]);
}

static bool
editFunctions(const char* ttl, bool ship, GSimObject& obj)
{
    WSimFunctionEditor fedit(ttl, ship, obj);
    ui_root->centerWidget(fedit);
    ui_root->add(fedit);
    int result = ui_root->run();
    ui_root->close(fedit);
    if (result == cm_Enter) {
        obj.setFlags(fedit.getNewFlags());
        return true;
    } else {
        return false;
    }
}

/** Edit special abilities of the specified simulator ship.
    \param sh Ship to edit
    \retval true Dialog was confirmed, ship possibly changed
    \retval false Dialog was cancelled, ship unchanged */
bool
editSimulatorShipFunctions(GSimShip& sh)
{
    return editFunctions(_("Ship Abilities"), true, sh);
}

/** Edit special abilities of the specified simulator planet.
    \param pl Planet to edit
    \retval true Dialog was confirmed, planet possibly changed
    \retval false Dialog was cancelled, planet unchanged */
bool
editSimulatorPlanetFunctions(GSimPlanet& pl)
{
    return editFunctions(_("Planet Abilities"), false, pl);
}

string_t
describeFunctions(const GSimObject& sh)
{
    string_t result;
    for (int i = 0; i < NUM_FUNCTIONS; ++i) {
        if (sh.hasFunction(GSimFunction(i))) {
            if (!result.empty()) {
                result += ", ";
            }
            result += _(LABELS[i]);
        }
    }
    if (result.empty()) {
        result = _("none");
    }
    return result;
}
