/**
  *  \file client/tiles/shipoverview.cc
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/shipoverview.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "game/ship.h"
#include "game/univ.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/mission.h"
#include "game/racename.h"
#include "util/misc.h"
#include "game/config.h"

/*************************** WShipOverviewTile **************************/

WShipOverviewTile::WShipOverviewTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(font_sizes[FONT_NORMAL_BOLD].scaledBy(30, 11), selection)
{ }

WShipOverviewTile::~WShipOverviewTile()
{ }

void
WShipOverviewTile::drawData(GfxCanvas& can)
{
    GShip* sh = dynamic_cast<GShip*>(getCurrentObject());
    if (!sh)
        return;

    GUniverse* univ = getCurrentUniverse();
    if (!univ)
        return;

    GShip& ps = *sh;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();

    int y = 0, x = 0;
    const int w = getInnerSize(0, 0).w;
    const int h = font_heights[FONT_NORMAL];

    // Line 1: Hull
    string_t line;
    if (ps.getRealOwner() != ps.getOwner()) {
        line += player_racenames.getAdjName(ps.getRealOwner());
        line += ' ';
    }
    int16_t level = ps.unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
    if (level >= 0) {
        line += getExperienceLevelName(level);
        line += ' ';
    }
    line += ps.getHull().getName();
    outTextF(ctx, x, y, w, line);
    y += h;

    // Line 2: Location
    outTextF(ctx, x, y, w, _("Location: ") + univ->getLocationName(ps.getPos(), locs_WW | locs_Orbit | locs_Verbose));
    y += h;

    // Line 3: Waypoint
    outTextF(ctx, x, y, w, _("Waypoint: ") + univ->getLocationName(ps.getWaypoint(), locs_WW | locs_Verbose));
    y += h;

    // Line 4: Movement, speed
    if (ps.getWaypointDX() == 0 && ps.getWaypointDY() == 0) {
        line = _("at waypoint");
    } else {
        line = format(_("%.2f ly"), distFromDX(ps.getWaypointDX(), ps.getWaypointDY()));
        line += ", ";
        if (ps.isHyperdriving()) {
            line += _("Hyperdrive");
        } else if (ps.getWarp() == 0) {
            line += _("not moving");
        } else {
            line += format(_("warp %d"), ps.getWarp());
        }
    }
    outTextF(ctx, x + h/2, y, w - h/2, "(" + line + ")");
    y += h;
    y += h/3;

    // Line 5: Mission
    if (const GMission* msn = GMissionList::getInstance().getMissionByShip(ps)) {
        line = msn->getLabel(ps);
    } else {
        line = format("M%d I%d T%d", ps.getMission(), ps.getInterceptId(), ps.getTowId());
    }
    outTextF(ctx, x, y, w, format(_("Mission: %s"), line));
    y += h;

    // Line 6: PE
    if (ps.getPrimaryEnemy() == 0) {
        line = _("none");
    } else {
        line = player_racenames.getShortName(ps.getPrimaryEnemy());
    }
    outTextF(ctx, x, y, w, format(_("Primary Enemy: %s"), line));
    y += h;
    y += h/3;

    // Line 7: Cargo [cut&paste from planet]
    outTextF(ctx, x, y, 50, _("Cargo:"));
    ctx.setTextAlign(2, 0);
    outTextF(ctx, x+130, y, 80, format("%dN", numToString(ps.getCargo(el_Neutronium))));
    outTextF(ctx, x+210, y, 80, format("%dT", numToString(ps.getCargo(el_Tritanium))));
    outTextF(ctx, x+290, y, 80, format("%dD", numToString(ps.getCargo(el_Duranium))));
    outTextF(ctx, x+370, y, 80, format("%dM", numToString(ps.getCargo(el_Molybdenum))));
    y += h;
    ctx.setTextAlign(0, 0);

    // Line 8: More cargo
    outTextF(ctx, x, y, w, format(_("%d colonist%!1{s%}, %d mc, %d suppl%1{y%|ies%}"),
                                  clansToString(ps.getCargo(el_Colonists)),
                                  numToString(ps.getCargo(el_Money)),
                                  numToString(ps.getCargo(el_Supplies))));
    y += h;
    y += h/3;

    // Line 9: FCode
    outTextF(ctx, x, y, w, format(_("FCode: %s"), ps.getFCode()));
    y += h;

    // Line 11: Misc
    if (ps.isFleetLeader()) {
        line = format(_("Leader of fleet #%d"), ps.getFleetNumber());
    } else if (ps.isFleetMember()) {
        line = format(_("Member of fleet #%d"), ps.getFleetNumber());
    } else if (ps.getRealOwner() != ps.getOwner()) {
        line = format(_("a %s ship"), player_racenames.getAdjName(ps.getRealOwner()));
    } else {
        line.clear();
    }
    outTextF(ctx, x, y, w, line);
}
