/**
  *  \file client/widgets/allylist.cc
  *  \brief Alliance Status Widget
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/allylist.h"
#include "game/alliance.h"
#include "game/racename.h"
#include "game/struct.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/nls.h"
#include "util/strtab.h"

static const char*const STATUS_LABELS[] = {
    N_("none"),
    N_("they offered"),
    N_("we offered"),
    N_("established"),
    N_("enemy"),
};

/** Constructor.
    \param id Widget Id
    \param player Player whose alliances we are editing
    \param editAllies Alliance configuration */
WAllyList::WAllyList(int id, int player, GAlliances& editAllies)
    : UIWidget(id),
      mouseDown(false),
      player(player),
      position(player == 1 ? 2 : 1),
      rightWidth(10 + getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(STATUS_LABELS, countof(STATUS_LABELS)))),
      editAllies(editAllies)
{ }

void
WAllyList::getLayoutInfo(LayoutInfo& info)
{
    // Reserve 15 ems for race names.
    // Right width is exact.
    // Add 10 pixels for chrome in left half, 10 pixels in right half (in constructor).
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(15, NUM_PLAYERS).movedBy(10 + rightWidth, 0);
    info.flags = li_FlexibleH;
}

void
WAllyList::drawContent(GfxCanvas& can)
{
    // Context
    GfxContext ctx;
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useCanvas(can);

    // Position
    const GfxRect ext = getExtent();
    const int listX      = ext.x;
    const int listWidth  = ext.w > rightWidth ? ext.w - rightWidth : 1;
    const int rightX     = ext.x + listWidth;
    const int rightWidth = ext.w - listWidth;
    const int lineHeight = font_heights[FONT_NORMAL];

    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        // We're modelling the listbox ColorStyle here
        UISkin lineSkin(i == position
                        ? getFocusState() == NoFocus
                        ? BRIGHT_GRAY_COLOR_SET
                        : BLACK_COLOR_SET
                        : GRAY_COLOR_SET);
        ctx.useColorScheme(lineSkin);
        ctx.setColor(i == player ? UIColor::tc_Faded : UIColor::tc_Static);

        int y = ext.y + lineHeight * (i-1);

        // Draw the race name
        GfxRect listExt(listX, y, listWidth, lineHeight);
        drawBackground(ctx, listExt);
        listExt.grow(-5, 0);
        ctx.setTextAlign(0, 0);
        outTextF(ctx, listExt, player_racenames.getShortName(i));

        // Figure out status
        bool weOffer   = editAllies.isAny(i, GAllianceLevel::IsOffer, true);
        bool theyOffer = editAllies.isAny(i, GAllianceLevel::IsOffer, false);
        bool enemy     = editAllies.isAny(i, GAllianceLevel::IsEnemy, true);

        uint8_t leftColor, rightColor, textColor;
        const char* text;
        if (i == player) {
            leftColor = rightColor = textColor = COLOR_SHIELD+4;
            text = "";
        } else if (weOffer) {
            if (theyOffer) {
                leftColor = COLOR_GREENSCALE + 9;
                textColor = COLOR_WHITE;
                text = _(STATUS_LABELS[3]);
            } else {
                leftColor = COLOR_YELLOW;
                textColor = COLOR_BLACK;
                text = _(STATUS_LABELS[2]);
            }
            rightColor = enemy ? uint8_t(COLOR_RED) : leftColor;
        } else {
            if (theyOffer) {
                leftColor = COLOR_YELLOW;
                rightColor = enemy ? uint8_t(COLOR_RED) : leftColor;
                textColor = COLOR_BLACK;
                text = _(STATUS_LABELS[1]);
            } else {
                leftColor = rightColor = enemy ? COLOR_RED : COLOR_FIRE+5;
                textColor = COLOR_YELLOW;
                text = enemy ? _(STATUS_LABELS[4]) : _(STATUS_LABELS[0]);
            }
        }

        // Draw the status
        GfxRect statusExt(rightX, y, rightWidth, lineHeight);
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        if (leftColor == rightColor) {
            drawSolidBar(ctx, statusExt, leftColor);
        } else {
            int split = (statusExt.w + statusExt.h)/2;
            int x = statusExt.x;
            int y = statusExt.y;
            for (int dy = 0; dy < statusExt.h; ++dy) {
                ctx.setColor(leftColor);
                drawHLine(ctx, x,         y, x + split-1);
                ctx.setColor(rightColor);
                drawHLine(ctx, x + split, y, x + statusExt.w-1);
                --split;
                ++y;
            }
        }
        ctx.setColor(textColor);
        ctx.setTextAlign(1, 0);
        outTextF(ctx, statusExt, text);
        drawFrameUp(can, statusExt);
    }

    can.update(getExtent());
}

bool
WAllyList::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && hasState(st_Focused)) {
        switch (event.key.code) {
         case SDLK_UP:
         case vk_WheelUp:
            setSelectedWidget();
            changePosition(-1);
            return true;

         case SDLK_DOWN:
         case vk_WheelDown:
            setSelectedWidget();
            changePosition(+1);
            return true;

         case SDLK_HOME:
         case SDLK_PAGEUP:
            setSelectedWidget();
            setPosition(player == 1 ? 2 : 1);
            return true;

         case SDLK_END:
         case SDLK_PAGEDOWN:
            setSelectedWidget();
            setPosition(player == NUM_PLAYERS ? NUM_PLAYERS-1 : NUM_PLAYERS);
            return true;

         case ' ':
            setSelectedWidget();
            toggleCurrent();
            return true;
        }
    }
    if (event.type == ev_Mouse) {
        if (isInside(GfxPoint(event.mouse.x, event.mouse.y))) {
            setSelectedWidget();
            int line = 1 + (event.mouse.y - getExtent().y) / font_heights[FONT_NORMAL];
            if (event.mouse.b) {
                setFocusedWidget();
                mouseDown = true;
                setPosition(line);
            } else {
                const int listWidth = getExtent().w > rightWidth ? getExtent().w - rightWidth : 1;
                if (mouseDown && event.mouse.x >= getExtent().x + listWidth) {
                    if (line == position) {
                        toggleCurrent();
                    }
                }
                mouseDown = false;
            }
            return true;
        } else {
            mouseDown = false;
            // Do not eat this event.
        }
    }
    return UIWidget::handleEvent(event, second_pass);
}

void
WAllyList::onStateChange(int state, bool enable)
{
    UIWidget::onStateChange(state, enable);
    if (state & st_Focused) {
        // Must redraw if focus changed
        drawWidget();
    }
    if (state & st_Selected) {
        // Forget that mouse is down when user moves it out
        mouseDown = false;
    }
}

/** Get position in list (player number). */
int
WAllyList::getPosition() const
{
    return position;
}

/** Set position in list (player number).
    \param newPos New position [1,NUM_PLAYERS] */
void
WAllyList::setPosition(int newPos)
{
    // Position must be valid and different from current
    if (newPos != player && newPos >= 1 && newPos <= NUM_PLAYERS && newPos != position) {
        position = newPos;
        drawWidget();
        sig_player_change.raise(position);
    }
}

/** Change position (player number).
    \param delta Direction (+1 or -1). */
void
WAllyList::changePosition(int delta)
{
    int newPos = position + delta;
    if (newPos == player) {
        newPos += delta;
    }
    setPosition(newPos);
}

/** Handle click on current line's right side.
    This offers/retracts an alliance. */
void
WAllyList::toggleCurrent()
{
    editAllies.setAll(position, GAllianceLevel::IsOffer, !editAllies.isAny(position, GAllianceLevel::IsOffer, true));
}
