/**
  *  \file client/widgets/combatdiagram.h
  *  \brief Combat Diagram Widget
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_WIDGETS_COMBATDIAGRAM_H
#define PCC_V2_CLIENT_WIDGETS_COMBATDIAGRAM_H

#include <vector>
#include "ui/ui.h"
#include "ui/widgets/tooltip.h"

class GVcrDatabase;
class GVcrObject;
class ProgressMonitor;

/** Combat diagram. This class contains the logic to compute a combat diagram.
    It takes a GVcrDatabase and puts the participating units into groups.
    It does so by building a list of Item's, each of which describes one unit
    i.e. one line in the diagram. Those items are grouped in a way that units
    that fight each other (i.e. are located at the same place in space) appear
    at neigbouring positions, but still reflect approximate battle order. */
class WCombatDiagram {
 public:
    struct Item {
        bool planet;              ///< true iff this item describes a planet.
        int16_t id;               ///< Id of this object.
        int16_t group_id;         ///< Group Id.
        int16_t sequence;         ///< Uniquifier to make sort stable.
        int16_t appears_in;       ///< Number of one fight this unit appears in.
        int16_t appears_as;       ///< Slot in fight \c appears_in where the unit appears.
        Item(bool planet, int id, int group_id, int sequence, int appears_in, int appears_as)
            : planet(planet), id(id), group_id(group_id), sequence(sequence), appears_in(appears_in), appears_as(appears_as)
            { }
    };

    WCombatDiagram();
    void init(GVcrDatabase& db, ProgressMonitor& prog);

    int getNumLines() const;
    const Item& getLine(int i) const;
    int findObjectLine(const GVcrObject& obj) const;

 private:
    std::vector<Item> lines;
    std::vector<Item>::const_iterator findObject(const GVcrObject& obj) const;
    void renameGroup(int from, int to);
};

/** Combat diagram widget. Displays a WCombatDiagram. The WCombatDiagram is expected
    to be fixed and not change during lifetime of the widget.

    User options:
    - colors can be mapped by player (as in statistics) or by team
    - user can click a battle */
class WCombatDiagramWidget : public UIWidget {
 public:
    WCombatDiagramWidget(GVcrDatabase& db, const WCombatDiagram& diag);

    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
    bool handleEvent(const UIEvent& event, bool second_pass);
    void onStateChange(int state, bool enable);
    int getHoverSlot() const;

    bool getTeamColorState() const;
    void setTeamColorState(bool state);

    /** Battle selection. Called when a battle is clicked. */
    Signal0 sig_click_battle;

 private:
    GVcrDatabase& db;
    const WCombatDiagram& diag;
    int hover_slot;
    bool mdown;
    bool team_colors;
    UITooltipControl tt_ctl;
    bool tt_active;

    void setHoverSlot(int slot);
    GfxRect getSlotBoundingBox(int slot) const;
    string_t getObjectName(int line) const;
    int getColorForOwner(int owner) const;
};

#endif
