/**
  *  \file client/widgets/vcrinfomain.cc
  *  \brief WVcrInfoMain widget
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/vcrinfomain.h"
#include "client/scr-sim.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/sim.h"
#include "game/spec.h"
#include "game/vcrdb.h"
#include "ui/msgbox.h"
#include "ui/richdocument.h"
#include "ui/util.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/unicodechars.h"

static const char*const headings[] = {
    N_("Primary:"),
    N_("Secondary:"),
    N_("Tech level:"),
    N_("Mass:"),
    N_("Fuel:"),
    N_("Engines:"),
    N_("Crew:"),
    N_("Experience:"),
    N_("Shield:"),
    N_("Damage:"),
};

static int
identifyEngine(const GVcrObject& o, const int assumedHull, const bool withESB)
{
    return o.identifyEngine(assumedHull, withESB);
}

/** Generate ship information.
    \param doc [out] Information is written here
    \param o [in] Object
    \param assumedHull [in] Assume this hull type
    \param withESB [in] true iff ESB is active in this fight
    \param fullInfo [in] true to show more details */
static void
showShipInfo(RichDocument& doc, const GVcrObject& o, const int assumedHull, const bool withESB, const bool fullInfo)
{
    const GHull& h = getHull(assumedHull ? assumedHull : 1);
    const int x1 = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(headings, countof(headings))) + font_ems[FONT_NORMAL];
    const int x2 = x1 + 16*font_ems[FONT_NORMAL];

    doc.addAt(x1, RichText(_("Ship")).withAttribute(RichTextStyleAttribute::Underline));
    doc.addAt(x2, RichText(_("Hull (max)")).withAttribute(RichTextStyleAttribute::Underline));
    doc.addNewline();

    // Primary weapon
    doc.add(_(headings[0]));
    if (o.getBeamCount() == 0) {
        doc.addAt(x1, _("none"));
    } else {
        doc.addAt(x1, string_t(format("%d " UTF_TIMES " %s", o.getBeamCount(), getBeam(o.getBeamType()).getName())));
    }
    if (assumedHull > 0) {
        if (h.getMaxBeams() > 0) {
            doc.addAt(x2, string_t(format(_("%d beam%!1{s%}"), h.getMaxBeams())));
        } else {
            doc.addAt(x2, _("none"));
        }
    }
    doc.addNewline();

    // Secondary weapon
    doc.add(_(headings[1]));
    if (o.getBayCount() > 0) {
        doc.addAt(x1, string_t(format(_("%d fighter bay%!1{s%}"), o.getBayCount())));
    } else if (o.getTorpLauncherCount() > 0) {
        doc.addAt(x1, string_t(format("%d " UTF_TIMES " %s", o.getTorpLauncherCount(), getTorp(o.getTorpType()).getName())));
    } else {
        doc.addAt(x1, _("none"));
    }
    if (assumedHull > 0) {
        if (h.getNumBays() > 0) {
            doc.addAt(x2, string_t(format(_("%d fighter bay%!1{s%}"), h.getNumBays())));
        } else if (h.getMaxLaunchers() > 0) {
            doc.addAt(x2, string_t(format(_("%d launcher%!1{s%}"), h.getMaxLaunchers())));
        } else {
            doc.addAt(x2, _("none"));
        }
    }
    doc.addNewline();

    // Ammo + Cargo
    if (o.getBayCount() > 0) {
        doc.addAt(x1, string_t(format(_("%d fighter%!1{s%}"), numToString(o.getFighterCount()))));
    } else if (o.getTorpLauncherCount() > 0) {
        doc.addAt(x1, string_t(format(_("%d torpedo%!1{es%}"), numToString(o.getTorpCount()))));
    } else {
        // No ammo. Add space so that the line is not empty; if it were, RichDocument
        // would reserve only half a line.
        doc.addAt(x1, " ");
    }
    if (assumedHull > 0) {
        doc.addAt(x2, string_t(format(_("%d kt cargo"), numToString(h.getCargo()))));
    }
    doc.addNewline();

    // Crew
    doc.add(_(headings[6]));
    doc.addAt(x1, numToString(o.getCrew()));
    if (assumedHull > 0) {
        doc.addAt(x2, numToString(h.getCrew()));
    }
    doc.addNewline();
    if (fullInfo && (config.NumExperienceLevels() > 0 || o.getExperienceLevel() > 0)) {
        doc.add(_(headings[7]));
        doc.addAt(x1, getExperienceLevelName(o.getExperienceLevel()));
        doc.addNewline();
    }

    // Tech level
    doc.add(_(headings[2]));
    if (assumedHull > 0) {
        doc.addAt(x2, numToString(h.getTech()));
    }
    doc.addNewline();

    // Mass
    doc.add(_(headings[3]));
    doc.addAt(x1, string_t(format(_("%d kt"), numToString(o.getMass()))));
    if (assumedHull > 0) {
        doc.addAt(x2, string_t(format(_("%d kt"), numToString(h.getMass()))));
    }
    doc.addNewline();
    if (fullInfo) {
        doc.add(_(headings[8]));
        doc.addAt(x1, string_t(format("%d%%", o.getShield())));
        doc.addNewline();

        doc.add(_(headings[9]));
        doc.addAt(x1, string_t(format("%d%%", o.getDamage())));
        doc.addAt(x2, raceId(o.getOwner()) == 2 ? "150%" : "99%");
        doc.addNewline();
    }

    // Fuel
    doc.add(_(headings[4]));
    if (assumedHull > 0) {
        doc.addAt(x2, string_t(format(_("%d kt"), numToString(h.getFuel()))));
    }
    doc.addNewline();

    // Engines
    doc.add(_(headings[5]));
    if (int eng = identifyEngine(o, assumedHull, withESB)) {
        doc.addAt(x1, getEngine(eng).getName());
    } else {
        doc.addAt(x1, _("unknown"));
    }
    if (assumedHull > 0) {
        doc.addAt(x2, numToString(h.getNumEngines()));
    }
    doc.addNewline();
}

/** Add ship to simulation.
    \param ss Simulation state
    \param o Object to add
    \param assumedHull Assumed hull, can be 0 */
static void
addShipToSim(GSimState& ss, const GVcrObject& o, const int assumedHull, const bool withESB)
{
    // Add or replace ship
    GSimShip* sh = ss.getShipById(o.getId());
    if (!sh) {
        sh = &ss.getShip(ss.addShip());
    }

    // Configure ship
    sh->setId(o.getId());
    sh->setName(o.getName());
    sh->setFCode("?""?""?");
    sh->setDamage(o.getDamage());
    sh->setShield(o.getShield());
    sh->setOwner(o.getOwner());
    sh->setExperienceLevel(o.getExperienceLevel());
    sh->setFlags(0);
    sh->setFlakRatingOverride(0);
    sh->setFlakCompensationOverride(0);
    sh->setHull(assumedHull);                // must be before crew, mass, and weapons
    sh->setCrew(o.getCrew());
    sh->setMass(o.getMass());
    sh->setBeamType(o.getBeamType());
    sh->setBeamCount(o.getBeamCount());
    sh->setTorpType(o.getTorpType());
    sh->setTorpLauncherCount(o.getTorpLauncherCount());
    sh->setBayCount(o.getBayCount());
    sh->setAmmo(o.getTorpLauncherCount() > 0 ? o.getTorpCount() : o.getFighterCount());
    int en = identifyEngine(o, assumedHull, withESB);
    sh->setEngineType(en ? en : 1);
    sh->setAggressiveness(GSimShip::agg_Kill);
    sh->setInterceptId(0);
}

/******************************** Planets ********************************/

namespace {
    struct Interval {
        int min, max;

        Interval()
            : min(1), max(0)
            { }
        void include(int n);
        void set(int n);
        string_t toString(int low, int high);
    };

    struct PlanetInformation {
        bool have_base;
        int  tech_state;
        Interval pdef, sbdef, sbfighters;

        bool init(const GVcrObject& o);
    };
}

/** Include value in interval. */
void
Interval::include(int n)
{
    if (min <= max) {
        // Interval was not empty
        if (min > n)
            min = n;
        if (max < n)
            max = n;
    } else {
        // Interval was empty
        min = max = n;
    }
}

inline void
Interval::set(int n)
{
    min = max = n;
}

/** Format interval into string.
    \param low,high Maximum bounds known to user */
string_t
Interval::toString(int low, int high)
{
    // Force into range
    if (min < low)
        min = low;
    if (max > high)
        max = high;

    if (min > max) {
        // Interval is empty
        return string_t(_("impossible"));
    } else if (min == max) {
        // Single value
        return numToString(min);
    } else if (max == high && min != low) {
        // Up to maximum
        return ::format(UTF_GEQ " %d", numToString(min));
    } else if (min == low && max != high) {
        // Down to minimum
        return ::format(UTF_LEQ " %d", numToString(max));
    } else {
        // General range, or full interval
        return ::format("%d ... %d", numToString(min), numToString(max));
    }
}

bool
PlanetInformation::init(const GVcrObject& o)
{
    int mass = o.getMass();
    mass -= 100;

    have_base = false;
    tech_state = 0;

    /* Unarmed? */
    if (mass <= 0) {
        if (o.getBeamCount() > 0) {
            /* Huh? */
            return false;
        }
    }

    /* Derived information */
    /* Fighters  = Round(Sqrt(PD)) + BaseFighters
       BeamCount = Round(Sqrt((PD+BD)/3))
       BeamType  = Round(Sqrt(PD/2)), at least BeamTech */
    /* Validate beam count */
    int beamCount = roundToInt(std::sqrt(mass / 3.0));
    if (beamCount <= 10 && o.getBeamCount() != beamCount) {
        return false;
    }

    /* Assume no starbase, and compute effective beam type and fighter count.
       If they do not match, there is a base. */
    int beamType = roundToInt(std::sqrt(mass / 2.0));
    if (beamType > 10) {
        beamType = 10;
    }
    int fighterCount = roundToInt(std::sqrt(double(mass)));
    int actualFighterCount = o.getFighterCount();

    /* FIXME: Nu uses different formulas.
         fighterCount = roundToInt(std::sqrt(mass - 0.75))    // or ERND?
         bayCount     = trunc(std::sqrt(mass)) */

    if (o.getBeamType() != beamType || actualFighterCount != o.getBayCount() || o.getBayCount() != fighterCount) {
        /* Initialize */
        have_base = true;
        tech_state = 0;

        /* We're pretty certain that there is a base. Try to figure out
           the PD/SBD split. PCC 1.x tried to compute the values directly,
           but didn't catch all cases. So we just use brute force now.
           It takes about 40k cycles maximum. */
        int maxSBD = config.MaximumDefenseOnBase(o.getOwner());
        int maxSBFighters = config.MaximumFightersOnBase(o.getOwner());
        for (int sbd = 0; sbd <= mass && sbd <= maxSBD; ++sbd) {
            // Compute derived information
            int pd = mass - sbd;
            int bt = roundToInt(std::sqrt(pd / 2.0));
            if (bt > 10) {
                bt = 10;
            }
            int ftr = roundToInt(std::sqrt(double(pd)));

            // Is this possible?
            // - Must not exceed reported type/count
            // - Consider that bases sometimes add 5 bays, sometimes not
            if (bt <= o.getBeamType()
                && ftr <= actualFighterCount
                && (ftr == o.getBayCount() || ftr + 5 == o.getBayCount())
                && actualFighterCount - ftr <= maxSBFighters)
            {
                if (bt < o.getBeamType()) {
                    tech_state |= 1;
                } else {
                    tech_state |= 2;
                }
                pdef.include(pd);
                sbdef.include(sbd);
                sbfighters.include(actualFighterCount - ftr);
            }
        }
        return true;
    } else {
        /* No hint that this might be a starbase */
        have_base = false;
        pdef.set(mass);
        tech_state = 0;
        sbdef.set(0);
        sbfighters.set(0);
        return true;
    }
}

static const char*const planetHeadings[] = {
    N_("Combat Mass:"),
    N_("Starbase:"),
    N_("Planetary Defense:"),
    N_("Starbase Defense:"),
    N_("Starbase Fighters:"),
    N_("Beam Tech:"),
};

/** Generate planet information.
    \param doc [out] Information is written here
    \param o [in] Object */
static void
showPlanetInfo(RichDocument& doc, const GVcrObject& o)
{
    int mass = o.getMass();

    const int x0 = font_ems[FONT_NORMAL]*2;
    const int x1 = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(planetHeadings, countof(planetHeadings))) + x0 + font_ems[FONT_NORMAL];

    /* Initial information */
    doc.add(_(planetHeadings[0]));
    doc.addAt(x1, string_t(format(_("%d kt"), numToString(mass))));
    doc.addParagraph();

    mass -= 100;
    if (mass <= 0) {
        doc.add(_("This planet is unarmed."));
        return;
    }

    doc.add(_("Derived Information:"));
    doc.addNewline();
    doc.setLeftMargin(x0);

    /* Compute derived information */
    PlanetInformation pi;
    if (!pi.init(o)) {
        doc.add(RichText(UIColor::tc_Red, _("Formula error.")));
        return;
    }

    doc.add(_(planetHeadings[1]));
    doc.addAt(x1, pi.have_base ? _("yes") : _("no"));
    doc.addNewline();

    doc.add(_(planetHeadings[2]));
    doc.addAt(x1, pi.pdef.toString(0, 10000));
    doc.addNewline();

    if (pi.have_base) {
        /* Base present: additional info */
        int maxSBD = config.MaximumDefenseOnBase(o.getOwner());
        int maxSBFighters = config.MaximumFightersOnBase(o.getOwner());

        doc.add(_(planetHeadings[3]));
        doc.addAt(x1, pi.sbdef.toString(0, maxSBD));
        doc.addNewline();

        doc.add(_(planetHeadings[4]));
        doc.addAt(x1, pi.sbfighters.toString(0, maxSBFighters));
        doc.addNewline();

        doc.add(_(planetHeadings[5]));
        if (pi.tech_state == 1) {
            // We know its the tech level which sets the value
            doc.addAt(x1, numToString(o.getBeamType()));
        } else if (pi.tech_state == 2 && o.getBeamType() != 10) {
            // We know its the DPs which set the value, so tech is less
            doc.addAt(x1, string_t(format(UTF_LEQ " %d", o.getBeamType())));
        } else {
            doc.addAt(x1, _("unknown"));
        }
        doc.addNewline();
    }
}

/** Add planet to simulation.
    \param ss Simulator state
    \param vcr Planet to add (before or after fight) */
static void
addPlanetToSim(GSimState& ss, const GVcrObject& o)
{
    /* Use /e/ to find out the planet's initial spec. Then, use /vcr/
       to generate the simulation participant.

       FIXME: damage handling is non-existant. We would have to scale
       planet/base defense somehow, but this differs in HOST and PHost. */
    PlanetInformation pi;
    if (!pi.init(o)) {
        messageBox(_("I was unable to figure out this planet's equipment."),
                   _("Add to Simulation"));
        return;
    }

    // Add planet
    ss.addPlanet();
    GSimPlanet& sp = ss.getPlanet();

    // Configure it
    sp.setId(o.getId());
    sp.setName(o.getName());
    sp.setFCode("ATT");
    sp.setDamage(o.getDamage());
    sp.setShield(o.getShield());
    sp.setOwner(o.getOwner());
    sp.setExperienceLevel(o.getExperienceLevel());
    sp.setFlags(0);
    sp.setFlakRatingOverride(0);
    sp.setFlakCompensationOverride(0);

    // We use the maximum possible planet defense, and the minimum
    // possible starbase defense. The rest follows from that.
    sp.setDefense(pi.pdef.max);
    if (pi.have_base) {
        // Defense, beams
        sp.setBaseDefense(pi.sbdef.min);
        int bt = roundToInt(std::sqrt(pi.pdef.max / 2.0));
        if (bt < o.getBeamType()) {
            sp.setBaseBeamTech(o.getBeamType());
        } else {
            sp.setBaseBeamTech(1);
        }

        // FIXME: starbase torps!
        sp.setBaseTorpTech(1);
        sp.setBaseTorps(1, 0);

        // Fighters. When using "after" data, starbase fighters
        // might have been destroyed.
        int ftr = o.getFighterCount() - roundToInt(std::sqrt(double(pi.sbdef.min)));
        if (ftr < 0) {
            sp.setBaseFighters(0);
        } else {
            sp.setBaseFighters(ftr);
        }
    }
}

/****************************** WVcrInfoMain *****************************/

WVcrInfoMain::WVcrInfoMain(bool fullInfo)
    : pEntry(0),
      side(0),
      hull(0),
      fullInfo(fullInfo)
{ }

WVcrInfoMain::~WVcrInfoMain()
{ }

void
WVcrInfoMain::drawContent(GfxCanvas& can)
{
    // Create blank document
    RichDocument doc;
    doc.setPageWidth(getExtent().w);

    // Generate info
    if (pEntry != 0) {
        const GVcrObject& obj = pEntry->getObject(side, false);
        if (obj.isPlanet()) {
            // Planet info
            showPlanetInfo(doc, obj);
        } else {
            // Ship info
            showShipInfo(doc, obj, hull, pEntry->isESBActive(), fullInfo);
        }
    }

    // Finish and draw
    doc.finish();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, getExtent());

    doc.draw(ctx, getExtent(), 0);
    can.update(getExtent());
}

void
WVcrInfoMain::getLayoutInfo(LayoutInfo& info)
{
    int lines = fullInfo ? 12 : 10;
    info.pref_size = info.min_size = font_sizes[FONT_NORMAL].scaledBy(35, lines);
    info.flags = 0;
}

void
WVcrInfoMain::set(GVcrEntry& entry, int side, int hull)
{
    this->pEntry = &entry;
    this->side   = side;
    this->hull   = hull;
    drawWidget();
}

void
WVcrInfoMain::addToSim(bool afterFight)
{
    // Anything selected?
    if (pEntry == 0) {
        return;
    }

    GSimState& ss = WSimScreen::sim_state;
    if (afterFight) {
        // Requested after fight, so first run the fight
        ProgressMonitor pm;
        pEntry->prepareResult(GVcrEntry::NeedCompleteResult | GVcrEntry::NeedQuickOutcome, pm);

        // Did we actually run the fight?
        if (pEntry->getPlayability() != GVcrEntry::IsPlayable) {
            messageBox(_("This fight could not be played. You can use [Shift+Ins] to use this unit's "
                         "status before the fight for simulation."),
                       _("Add to Simulation"));
            return;
        }

        // Did the unit survive the fight?
        if (pEntry->getOutcome(side) < 0) {
            messageBox(_("This unit did not survive the fight. You can use [Shift+Ins] to use this unit's "
                         "status before the fight for simulation."),
                       _("Add to Simulation"));
            return;
        }
    }

    // Add it
    const GVcrObject& obj = pEntry->getObject(side, afterFight);
    if (obj.isPlanet()) {
        addPlanetToSim(ss, obj);
    } else {
        addShipToSim(ss, obj, hull, pEntry->isESBActive());
    }

    ss.doUpdates();
}
