/**
  *  \file game/ship-form.cc
  *  \brief Ship Formulas and Prediction
  *
  *  This module contains functions to compute ship properties and
  *  predict their behaviour.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <cmath>
#include "game/ship.h"
#include "game/mission.h"
#include "game/player.h"
#include "game/config.h"
#include "game/coord.h"
#include "game/planet.h"
#include "game/formula.h"
#include "game/host.h"
#include "game/univ.h"
#include "game/fcode.h"
#include "game/team.h"
#include "game/reg.h"
#include "util/misc.h"
#include "util/recode.h"
#include "game/shippredict.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "util/unicodechars.h"

static int
getMovementTurns(GUniverse& univ, int sid, GPoint pos, GPoint waypoint, int warp)
{
    GShipTurnPredictor pred(univ, sid);
    pred.setPosition(pos);
    pred.setWaypoint(waypoint);
    pred.setSpeed(warp);
    pred.computeMovement();
    return pred.getNumTurns();
}

/** Set ship to its optimum warp. */
void
setOptimumWarp(GUniverse& univ, int sid)
{
    // FIXME: I'd like this to take a 'const GShip', but we need univ/sid for
    // GShipTurnPredictor, and we need to be able to modify it as ships aren't copyable.
    GShip& sh = univ.getShip(sid);

    /* sort out trivial cases */
    if (sh.isHyperdriving() || (sh.getWaypointDX() == 0 && sh.getWaypointDY() == 0))
        return;

    int n = computeOptimumWarp(univ, sid, sh.getPos(), sh.getWaypoint());
    if (n != 0) {
        sh.setWarp(mp16_t(n));
    }
}

int
computeOptimumWarp(GUniverse& univ, int sid, GPoint pos, GPoint waypoint)
{
    /* Find out how long it takes to get there at this speed */
    int this_speed = getEngine(univ.getShip(sid).getEngineType()).getMaxEfficientWarp();
    int result = this_speed;
    int this_time = getMovementTurns(univ, sid, pos, waypoint, this_speed);

    /* Do we make it in finite time? If not, let user resolve it. */
    if (this_time >= GShipTurnPredictor::MOVEMENT_TIME_LIMIT)
        return result;

    /* Find lowest possible speed.
       - into deep space: allow all speeds down to warp 1
       - into warp well: assume they want to reach the planet. Do not go below warp 2.
       - stay in same warp well: allow all speeds down to warp 1 */
    int old_planet = univ.getPlanetAt(pos, true);
    int new_planet = univ.getPlanetAt(waypoint, true);
    int lower_limit;
    if (new_planet != 0 && old_planet != new_planet) {
        lower_limit = 2;
    } else {
        lower_limit = 1;
    }

    /* Find whether we can reach the target with a slower speed */
    while (this_speed > lower_limit) {
        --this_speed;
        if (getMovementTurns(univ, sid, pos, waypoint, this_speed) > this_time) {
            /* It's slower, so undo and stop */
            return this_speed+1;
        }
    }
    return this_speed;
}

/****************************** Mine Laying ******************************/

/** Parse mine laying mission.
    \return true if ship has a mine laying mission */
bool
parseMineLayingMission(const GShip& sh, const GUniverse& univ, GMineMissionData& md)
{
    /* Check whether all required values are known */
    if (!sh.getMission().isKnown()
        || !sh.getTowId().isKnown()
        || !sh.getInterceptId().isKnown()
        || !sh.getTorpType().isKnown()
        || !sh.getNumTorpLaunchers().isKnown()
        || !sh.getAmmo().isKnown())
    {
        return false;
    }

    /* Check whether ship can lay mines */
    int ttyp  = sh.getTorpType();
    int tlaun = sh.getNumTorpLaunchers();
    int torps = sh.getAmmo();
    if (ttyp <= 0 || tlaun <= 0 || torps <= 0) {
        return false;
    }

    /* Check mission */
    bool usemix = false, usemdx = false;     // Use mix/mdx fcode?
    bool makeweb = false;                    // Make a web field?
    int race = 0;                            // Lay as this race?
    int reqid = 0;                           // Require this Id?
    int torplimit = 0;                       // Torpedo limit
    const int owner = sh.getRealOwner();
    const int mission = sh.getMission();
    if (mission == GMission::msn_LayMines) {
        /* Lay mines: honors mix, mdx, but no other parameters */
        usemix = usemdx = true;
    } else if (mission == GMission::msn_Special) {
        /* Lay web (race is checked later): honors, mix, mdx, but no other parameters */
        usemix = usemdx = true;
        makeweb = true;
    } else if (!host.isPHost() || !allowRegistered(owner)) {
        /* Following missions only for registered players in PHost */
        /* FIXME: as for fcodes, we could query mission.cc for presence of these missions */
        return false;
    } else if (mission == config.ExtMissionsStartAt() + GMission::pmsn_LayMines) {
        /* Lay Minefield: honors no fcodes, registered-only */
        torplimit = sh.getInterceptId();
        race      = sh.getTowId();
    } else if (mission == config.ExtMissionsStartAt() + GMission::pmsn_LayWeb) {
        /* Lay Web Minefield: honors no fcodes, registered-only */
        torplimit = sh.getInterceptId();
        race      = sh.getTowId();
        makeweb   = true;
    } else if (mission == config.ExtMissionsStartAt() + GMission::pmsn_LayMinesIn) {
        /* Lay Mines In: honors miX */
        usemix    = true;
        torplimit = sh.getInterceptId();
        reqid     = sh.getTowId();
    } else if (mission == config.ExtMissionsStartAt() + GMission::pmsn_LayWebIn) {
        /* Lay Web Mines In: honors miX */
        usemix    = true;
        torplimit = sh.getInterceptId();
        reqid     = sh.getTowId();
        makeweb   = true;
    } else {
        /* Not a mine laying mission */
        return false;
    }

    /* Postprocess fcodes */
    bool used_fc = false;
    if (usemix || usemdx) {
        /* We accept a friendly code if it
               - is not listed in fcodes.cc (assuming that is blank, missing, unmaintained)
           OR  -     works on the ship
                 AND is not restricted to registered ships, or player is registered */
        const string_t       fc  = sh.getFCode();
        const GFCodeList&    fcl = GFCodeList::getInstance();
        GFCodeList::iterator fci = fcl.getFCodeByName(fc);
        bool validfc = (fci == fcl.end()
                        || (fci->worksOn(sh)
                            && ((fci->getFlags() & GFCode::fc_Registered) == 0
                                || allowRegistered(owner))));
        if (validfc && fc.size() == 3) {
            if (usemix && fc[0] == 'm' && fc[1] == 'i') {
                /* miX */
                race = getPlayerIdFromChar(fc[2]);
                used_fc = (race > 0);
            } else if (usemdx && fc[0] == 'm' && fc[1] == 'd') {
                /* mdX */
                if (fc[2] == '0')
                    torplimit = 100, used_fc = true;
                else if (fc[2] >= '1' && fc[2] <= '9')
                    torplimit = 10*(fc[2]-'0'), used_fc = true;
                else if (fc[2] == 'h')
                    torplimit = torps/2, used_fc = true;
                else if (fc[2] == 'q')
                    torplimit = torps/4, used_fc = true;
            }
        }
    }

    /* Postprocess race and limit */
    if (race <= 0 || race > NUM_OWNERS)
        race = owner;
    if (torplimit > 0 && torplimit < torps)
        torps = torplimit;

    /* Check whether this type is allowed */
    if (makeweb) {
        if (raceMissionId(owner) != 7 || !config.AllowWebMines())
            return false;
    } else {
        if (!config.AllowMinefields())
            return false;
    }

    /* Postprocess race, reqid */
    const GMinefieldType& mfc = univ.ty_minefields;
    if (reqid != 0) {
        /* We're requiring a particular minefield, so check whether it exists
           and matches our parameters. If it doesn't, we refuse to lay mines,
           in the same way PHost does. */
        GMinefield* mf = mfc.get(reqid);
        if (!mf)
            return false;

        if (host.hasAutomaticMineIdentity())
            race = mf->getOwner();

        if (mf->getOwner() != race
            || mf->isWeb() != makeweb
            || distanceSquared(mf->getPos(), sh.getPos()) > mf->getUnitsForLaying())
        {
            return false;
        }
    } else {
        /* We're not requiring a particular minefield, so find one. */
        long closest = 0;
        if (host.hasMinefieldCenterBug()) {
            for (GMinefieldType::const_iterator i = mfc.begin(); i != mfc.end(); ++i) {
                if (i->getOwner() == race && i->isWeb() == makeweb) {
                    long dist = distanceSquared(i->getPos(), sh.getPos());
                    if (reqid == 0 || dist < closest) {
                        /* Minefield matches type and is close. We note its
                           Id only when we're inside; set it to zero if we're
                           outside the field and make a new one */
                        reqid = (dist <= i->getUnitsForLaying() ? i->getId() : 0);
                        closest = dist;
                    }
                }
            }
        } else {
            for (GMinefieldType::const_iterator i = mfc.begin(); i != mfc.end(); ++i) {
                if (i->getOwner() == race && i->isWeb() == makeweb) {
                    long dist = distanceSquared(i->getPos(), sh.getPos());
                    if (dist <= i->getUnitsForLaying() && (reqid == 0 || dist < closest)) {
                        /* Minefield matches type and is close, and we're inside */
                        reqid = i->getId();
                        closest = dist;
                    }
                }
            }
        }
    }

    /* Find mine laying rate */
    int rate, srate, mrate;
    if (makeweb) {
        srate = config.UnitsPerWebRate(owner);
        mrate = config.UnitsPerWebRate(race);
    } else {
        srate = config.UnitsPerTorpRate(owner);
        mrate = config.UnitsPerTorpRate(race);
    }
    rate = (srate < mrate ? srate : mrate);

    /* We now know that we're laying mines, so fill in the structure */
    md.mine_id   = reqid;
    md.owner     = race;
    md.type      = makeweb;
    md.num_torps = torps;
    md.num_units = int32(torps) * rate * square(sh.getTorpType()) / 100;
    md.used_mission = true;
    md.used_fcode = used_fc;
    return true;
}

bool
parseMineScoopMission(const GShip& sh, const GUniverse& /*univ*/, GMineMissionData& md)
{
    /* Check whether all required values are known */
    if (!sh.getMission().isKnown()
        || !sh.getTowId().isKnown()
        || !sh.getInterceptId().isKnown()
        || !sh.getTorpType().isKnown()
        || !sh.getNumTorpLaunchers().isKnown()
        || !sh.getBeamType().isKnown()
        || !sh.getNumBeams().isKnown())
    {
        return false;
    }

    /* Check whether ship can scoop mines */
    if (sh.getTorpType() <= 0
        || sh.getNumTorpLaunchers() <= 0
        || (host.isPHost()
            && (sh.getBeamType() <= 0
                || sh.getNumBeams() <= 0)))
    {
        return false;
    }

    /* Do we want to scoop? */
    int mission = sh.getMission();
    if (mission == GMission::msn_MineSweep) {
        /* Check for "msc" fcode */
        const GFCodeList&    fcl = GFCodeList::getInstance();
        GFCodeList::iterator fci;
        if (sh.getFCode() == "msc"
            && ((fci = GFCodeList::getInstance().getFCodeByName(sh.getFCode())) == fcl.end()
                || (fci->worksOn(sh)
                    && ((fci->getFlags() & GFCode::fc_Registered) == 0
                        || allowRegistered(sh.getRealOwner())))))
        {
            /* accept */
            md.mine_id   = 0;
            md.owner     = sh.getRealOwner();
            md.type      = false; // irrelevant for scooping
            md.num_torps = 0;     // means: no limit
            md.num_units = 0;     // irrelevant for scooping
            md.used_mission = true;
            md.used_fcode = true;
            return true;
        } else {
            /* reject */
            return false;
        }
    } else if (host.isPHost() && mission == config.ExtMissionsStartAt() + GMission::pmsn_ScoopTorps) {
        /* PHost mission */
        md.mine_id   = sh.getTowId();
        md.owner     = sh.getRealOwner();
        md.type      = false;
        md.num_torps = sh.getInterceptId();
        md.num_units = 0;
        md.used_mission = true;
        md.used_fcode = false;
        return true;
    } else {
        return false;
    }
}

/******************************* Chunneling ******************************/

/** Check for possible chunnel failures.
    \param sh [in] Ship to check
    \param univ [in] Universe to check
    \param minFuel [in] Minimum fuel required on this ship */
static int
checkChunnelFailures(const GShip& sh, const GUniverse& univ, const int minFuel)
{
    int result = 0;
    if (host.isPHost() && sh.getDamage().isKnown() && sh.getDamage() >= config.DamageLevelForChunnelFail()) {
        result |= GChunnelData::chf_MateDamaged;
    }
    if (sh.getCargo(el_Neutronium).isKnown() && sh.getCargo(el_Neutronium) <= minFuel) {
        result |= GChunnelData::chf_MateFuel;
    }
    if (sh.getWarp().isKnown() && sh.getWarp() > 0) {
        result |= GChunnelData::chf_MateMoving;
    }
    if (univ.isShipTowed(sh.getId())) {
        result |= GChunnelData::chf_MateTowed;
    }
    return result;
}

static bool
isAlliedOrOwn(const GShip& sh, const GShip& mate)
{
    // mate.isPlayable(GObject::Playable)
    GShip::Kind mateKind = mate.getShipKind();
    return mate.isReliablyVisible(sh.getOwner())
        && (mateKind == GShip::CurrentShip || mateKind == GShip::CurrentTarget)
        && (mate.getOwner() == sh.getOwner()
            || (config.AllowAlliedChunneling()
                && getPlayerTeam(sh.getOwner()) == getPlayerTeam(mate.getOwner())));
}

/** Check for valid chunnel mate.
    \param initiator Chunnel initiator
    \param mate Potential chunnel mate */
bool
isValidChunnelMate(const GShip& initiator, const GShip& mate)
{
    int32_t dist = host.isPHost() ? square(config.MinimumChunnelDistance()) : 9901;
    return isAlliedOrOwn(initiator, mate)
        && mate.getFleetNumber() == 0
        && (mate.canDoSpecial(hf_FirecloudChunnel) || mate.canDoSpecial(hf_ChunnelTarget))
        && (distanceSquared(initiator.getPos(), mate.getPos().getSimpleNearestAlias(initiator.getPos())) >= dist);
}

/** Check chunnel mission.
    \param sh [in] Potential chunnel initiator
    \param univ [in] Universe
    \param cd [out] Chunnel info
    \return true iff ship is chunneling

    Postconditions:
    - if ship does not chunnel, ch.target is 0
    - if ship does chunnel, ch.target is ship id, and target ship does exist
      . then, ch.failure contains bitfield of reasons why chunnel fails
        (consequently, it's 0 to predict success)
      . ch.kind is the kind of chunnel, bitfield of ship classes to take with
        me (consequently, nonzero for a successful result)

    Originally shipacc.pas::CheckChunnel. */
bool
parseChunnelMission(const GShip& sh, const GUniverse& univ, GChunnelData& cd)
{
    cd.target  = 0;
    cd.failure = 0;
    cd.kind    = 0;

    /* For simplicity, check 3-place numeric FCs only */
    string_t fc = sh.getFCode();
    int sid;
    if (fc.size() == 3
        && stringValue(fc, sid)
        && univ.isValidShipId(sid)
        && sid != sh.getId())
    {
        const GShip& mate = univ.getShip(sid);
        if (isAlliedOrOwn(sh, mate) && (mate.canDoSpecial(hf_FirecloudChunnel) || mate.canDoSpecial(hf_ChunnelTarget))) {
            /* Target exists and can receive a chunnel. Can we start one?
               Note that we have rejected self-chunnel above. Hosts fail it
               implicitly for violating minimum distance. */
            if (sh.canDoSpecial(hf_FirecloudChunnel)) {
                cd.kind = cd.chk_Self | cd.chk_Others;
            } else {
                // Check both of the lesser abilities, ship may have both
                if (sh.canDoSpecial(hf_ChunnelSelf)) {
                    cd.kind |= cd.chk_Self;
                }
                if (sh.canDoSpecial(hf_ChunnelOthers)) {
                    cd.kind |= cd.chk_Others;
                }
            }

            /* Found? */
            if (cd.kind != 0) {
                cd.target = sid;

                /* Now figure out failure modes
                   - mate is moving/being towed/fuelless/excessively damaged
                   - we have < 51 fuel
                   - we are moving/being towed/excessively damaged
                   - minimum distance violated */
                int minFuel = host.getMinimumFuelToInitiateChunnel() - 1;
                cd.failure |= checkChunnelFailures(mate, univ, 0);
                cd.failure |= checkChunnelFailures(sh, univ, minFuel) * (cd.chf_Damaged/cd.chf_MateDamaged);
                if (host.isPHost() && sh.getMission().isKnown() && sh.getMission() == GMission::pmsn_Training + config.ExtMissionsStartAt()) {
                    cd.failure |= cd.chf_Training;
                }

                /* Distance check:
                   HOST uses    ERND(Sqrt(....)) >= 100, i.e.   Sqrt(....) >= 99.5
                   PHost uses   Trunc(Sqrt(....)) >= MCD, i.e.  Sqrt(....) >= MCD */
                int32_t dist = host.isPHost() ? square(config.MinimumChunnelDistance()) : 9901;
                if (distanceSquared(sh.getPos(), mate.getPos()) < dist) {
                    cd.failure |= cd.chf_Distance;
                }
            }
        }
    }

    return cd.kind != 0;
}

string_t
formatChunnelFailures(int failures)
{
    static const char*const problems[] = {
        N_("Initiator damaged"),
        N_("Initiator needs fuel"),
        N_("Initiator moving"),
        N_("Initiator under tow"),
        N_("Initiator is training"),
        N_("Mate damaged"),
        N_("Mate needs fuel"),
        N_("Mate moving"),
        N_("Mate under tow"),
        N_("Distance too short"),
    };
    static const int problemFlags[] = {
        GChunnelData::chf_Damaged,
        GChunnelData::chf_Fuel,
        GChunnelData::chf_Moving,
        GChunnelData::chf_Towed,
        GChunnelData::chf_Training,
        GChunnelData::chf_MateDamaged,
        GChunnelData::chf_MateFuel,
        GChunnelData::chf_MateMoving,
        GChunnelData::chf_MateTowed,
        GChunnelData::chf_Distance,
    };
    string_t probStr;
    for (uint_t i = 0; i < countof(problemFlags); ++i) {
        if (failures & problemFlags[i]) {
            probStr += UTF_BULLET " ";
            probStr += _(problems[i]);
            probStr += "\n";
        }
    }
    return probStr;
}


int
getTurnFuelUsage(const GShip& sh, const ConfigStdOption& opt, int turns)
{
    if (!sh.getHullId().isKnown()) {
        return 0;
    } else {
        return getTurnFuelUsage(sh.getHullId(), sh.getRealOwner(), opt, turns);
    }
}

int
getTurnFuelUsage(int hull, int player, const ConfigStdOption& opt, int turns)
{
    if (!host.isPHost()) {
        return 0;
    } else {
        int result = (getHull(hull).getMass() * opt(player) + 99) / 100;
        if (turns > 0) {
            result *= turns;
        }
        return result;
    }
}
