/**
  *  \file game/sim-run.h
  *  \brief Running Simulations
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SIM_RUN_H
#define PCC_V2_GAME_SIM_RUN_H

#include <vector>
#include "game/sim.h"
#include "cpluslib/types.h"
#include "cpluslib/sigslot.h"
#include "util/progress.h"
#include "game/vcrdb.h"

/*! \addtogroup Simulator */
/*@{*/

/** Result of a single simulation. In addition to a set of battles and
    an updated status, the simulator can return a weight. This is
    exclusively used in seed control mode. For example, when doing
    left/right balancing using Balance360k, the fight with bonus has a
    probability of 59%; the unmodified fight appears with 41%. In
    addition, appearance of this bonus increases series length from
    110 to 220.

    Before handing this to runSimulation(), call init(). */
struct GSimBattleResult {
    /** Weight of this battle for statistics purposes. Filled in by
        simulator. Called 'BattleWeight' in ccsim.pas. */
    int32 this_battle_weight;
    /** Total weight of a series. Filled in by simulator, must be
        constant for one set of options. Called 'Divider' in
        ccsim.pas. */
    int32 total_battle_weight;
    /** Length of a series. This is the number of possibly-different
        battles that can appear. Filled in by simulator, must be
        constant for one set of options. Called 'Modulus' in
        ccsim.pas. */
    int32 series_length;
    /** Index of this battle, 0-based. Filled in by driver code. Used
        by simulator to determine where in a non-equal set we are. */
    int32 this_battle_index;
    /** Actual battle. Filled in by simulator. */
    Ptr<GVcrDatabase> battles;

    /** Constructor. */
    GSimBattleResult(const GSimOptions& opts)
        { init(opts, 0); }

    void init(const GSimOptions& opts, int this_battle_index);
    int addSeries(int length);

    void changeWeightTo(int32 new_weight);
};

void runSimulation(GSimState& state, const GSimOptions& opts, GSimBattleResult& result, ProgressMonitor& monitor);

bool sortByBattleOrderTHost(const GSimShip* a, const GSimShip* b);
bool sortByBattleOrderPHost(GSimObject* a, GSimObject* b);

/*@}*/

#endif
