/**
  *  \file gfx/sprite3d.h
  *  \brief Sprite Management
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_SPRITE3D_H
#define PCC_V2_GFX_SPRITE3D_H

#include <vector>
#include "cpluslib/object.h"
#include "gfx/gfxutil.h"
#include "cpluslib/smartptr.h"
#include "gfx/gfx.h"
#include "gfx/pixmap.h"

class GfxSpriteManager3D;

///! Sprite in pseudo-3-D space.
class GfxSprite3D : public Object {
    friend class GfxSpriteManager3D;

    /** Data used for communication to GfxSpriteManager3D. */
    class ManagerData {
        friend class GfxSpriteManager3D;
        friend class GfxSprite3D;

        GfxSpriteManager3D* manager;
        bool need_update  : 1;    ///< True iff sprite needs to be redrawn.
        bool need_delete  : 1;    ///< True iff sprite asked to be deleted.
        bool was_moved    : 1;    ///< True iff sprite was moved.
        bool tick_mark    : 1;    ///< Used by GfxSpriteManager3D::tick().
        bool is_invisible : 1;    ///< True if sprite is currently invisible/off-screen.
        GfxRect old_area;         ///< Old position. Valid if was_moved is true.

        ManagerData()
            : manager(0), need_update(0), need_delete(0), was_moved(0),
              tick_mark(0), is_invisible(0)
            { }
    };

    /** Position in 2D area. */
    GfxRect pos_2d;
    /** Relative position in Z order. Nonzero values are visible, zero
        is topmost. */
    int pos_z;
    /** Identifier. */
    int identifier;
    /** Data related to sprite manager. */
    ManagerData md;

    void setManager(GfxSpriteManager3D* manager);
 public:
    /** Construct blank sprite. The sprite is at point (0,0,0) and has
        identifier zero. */
    GfxSprite3D()
        : pos_2d(0, 0, 0, 0), pos_z(0), identifier(0), md()
        { }
    /** Construct sprite at given position. The identifier will be zero.
        \param area screen position
        \param z_pos stacking position */
    GfxSprite3D(GfxRect area, int z_pos)
        : pos_2d(area), pos_z(z_pos), identifier(0), md()
        { }
    /** Construct sprite at given position.
        \param area screen position
        \param z_pos stacking position
        \param identifier identifier */
    GfxSprite3D(GfxRect area, int z_pos, int identifier)
        : pos_2d(area), pos_z(z_pos), identifier(identifier), md()
        { }
    /** Construct blank with given identifier. The sprite will be at
        point (0,0,0).
        \param identifier identifier */
    GfxSprite3D(int identifier)
        : pos_2d(0, 0, 0, 0), pos_z(0), identifier(identifier), md()
        { }

    virtual void tick();
    /** Draw this sprite on given canvas.
        \param can target canvas. Caller will lock/update it. */
    virtual void draw(GfxCanvas& can) = 0;

    /** Get this sprite's position on the canvas. */
    const GfxRect& getPos() const
        { return pos_2d; }
    /** Get this sprite's position in the Z order. */
    int getPosZ() const
        { return pos_z; }
    /** Get associated sprite manager. */
    GfxSpriteManager3D* getManager() const
        { return md.manager; }
    /** Get center X position. */
    int getPosCX() const
        { return pos_2d.x + pos_2d.w/2; }
    /** Get center Y position. */
    int getPosCY() const
        { return pos_2d.y + pos_2d.h/2; }
    /** Get identifier. */
    int getId() const
        { return identifier; }

    void deleteThis();
    void update();
    void setPos(GfxRect area, int z);
    void setPos2D(GfxPoint pt);
    void setId(int id);
    void setCenterPos2D(GfxPoint pt);
    void setCenterPos(GfxPoint pt, int z);
    void ensureTick();
};

//! Sprite Manager.
class GfxSpriteManager3D {
    friend class GfxSprite3D;

    typedef std::vector<Ptr<GfxSprite3D> > container_t;
    container_t   sprites;
    GfxRect       bbox;
    bool          need_sort : 1;     ///< True iff any sprite changed its Z or Id, so we need to sort.
    bool          need_delete : 1;   ///< True iff any sprite asked to be deleted.
    bool          need_update : 1;   ///< True iff any sprite wants to be redrawn.
    bool          tick_mark : 1;     ///< Current tick mark for tick().

    void drawBackground(GfxCanvas& can, GfxRect area);
 public:
    GfxSpriteManager3D(GfxRect bbox);
    ~GfxSpriteManager3D();

    void tick();
    void doUpdates(GfxCanvas& can);
    void doUpdatesSilently();
    void addSprite(Ptr<GfxSprite3D> sprite);
    void draw(GfxCanvas& can);

    void deleteSpritesInZRange(int z0, int z1);
    void deleteSpritesAfterZ(int z0);
    bool haveSpritesInZRange(int z0, int z1) const;
    void deleteSpritesWithId(int id);
    bool haveSpritesWithId(int id) const;

    /** Get current bounding box. \see setBBox */
    const GfxRect& getBBox() const
        { return bbox; }
    /** Set bounding box. All output will be clipped to that area.
        \param r area */
    void setBBox(const GfxRect& r)
        { bbox = r; }

    /** Delete all sprites at a specific Z coordinate.
        \param z Z coordinate */
    void deleteSpritesAtZ(int z)
        { deleteSpritesInZRange(z, z+1); }
    /** Check whether we have a sprite at a specific Z coordinate.
        \param z Z coordinate */
    bool haveSpritesAtZ(int z) const
        { return haveSpritesInZRange(z, z+1); }

    /** Iterator for processing sprites. */
    typedef container_t::const_iterator iterator;
    /** First sprite. */
    iterator begin() const
        { return sprites.begin(); }
    /** One-past-last sprite. */
    iterator end() const
        { return sprites.end(); }
};

//! Static pixmap sprite.
class GfxPixmapSprite3D : public GfxSprite3D {
    Ptr<GfxPixmap> pix;
 public:
    GfxPixmapSprite3D(int x, int y, int z, Ptr<GfxPixmap> pix);
    GfxPixmapSprite3D(int x, int y, int z, Ptr<GfxPixmap> pix, bool);
    GfxPixmapSprite3D(int x, int y, int z, int id, Ptr<GfxPixmap> pix);
    void draw(GfxCanvas& can);
};

#endif
