/**
  *  \file u/t_game_planetpred.cc
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/planetpred.h"

#include "u/t_game.h"
#include "game/planet.h"
#include "game/config.h"
#include "game/host.h"
#include "cpluslib/format.h"

namespace {
    const TPlanet PLANET_DATA = {
        1,                          // owner
        44,                         // id
        {'a','b','c'},              // fcode
        0, 0, 0,                    // mi/fa/de
        { 100, 100, 100, 100 },     // mined
        99999,                      // clans
        0,                          // supplies
        0,                          // money
        { 100, 100, 100, 100 },     // ground
        { 50, 50, 50, 50 },         // density
        0, 0,                       // ctax/ntax
        100, 100,                   // chap/nhap
        5,                          // ngov
        0,                          // natives
        9,                          // nrase
        0,                          // temp
        0,                          // build base
    };

    void runTest(const char* name, const TPlanet& data, const int32_t (&expect)[101])
    {
        for (int i = 0; i <= 100; ++i) {
            // Initialize planet
            GPlanet planet(data.planet_id);
            planet.setXY(1000, 1000);
            planet.addPlanetData(data, GPlayerSet(data.owner));
            planet.setTemperature(mp16_t(i));

            // Compute turn
            computeTurn(planet, GPlanetEffectors());

            // Check result
            string_t label = format("%s [%d deg]", name, i);
            TSM_ASSERT_EQUALS(label.c_str(), int32_t(planet.getColonists()), expect[i]);
        }
    }

    void runTestNatives(const char* name, const TPlanet& data, const int32_t (&expect)[101])
    {
        for (int i = 0; i <= 100; ++i) {
            // Initialize planet
            GPlanet planet(data.planet_id);
            planet.setXY(1000, 1000);
            planet.addPlanetData(data, GPlayerSet(data.owner));
            planet.setTemperature(mp16_t(i));

            // Compute turn
            computeTurn(planet, GPlanetEffectors());

            // Check result
            string_t label = format("%s [%d deg]", name, i);
            TSM_ASSERT_EQUALS(label.c_str(), int32_t(planet.getNatives()), expect[i]);
        }
    }
}


void
GamePlanetPredTestSuite::testGrowthMaxTim()
{
    static const int32_t EXPECT[101] = {
        99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999,
        45520, 48293, 51018, 53693, 56315, 58882, 61390, 63838, 66223, 68543, 70795, 72977, 75088, 77124, 79085,
        80967, 82770, 84491, 86128, 87681, 89148, 90526, 91815, 93014, 94121, 95135, 96056, 96881, 97611, 98245,
        98782, 99222, 99564, 99808, 99953,
        100000,
        99948, 99798, 99549, 99202, 98758,
        98216, 97577, 96842, 96011, 95086, 94067, 92955, 91752, 90458, 89075, 87605, 86047, 84405, 82680, 80874,
        78987, 77023, 74983, 72869, 70683, 68427, 66104, 63715, 61264, 58753, 56183, 53558, 50881, 48153, 99999,
        99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999, 99999
    };

    // Config
    config.assignDefaults();
    config.ClimateDeathRate.set(0);
    config.ClimateLimitsPopulation.set(1);

    // Host
    host.setHostVersion(MKVERSION(3,22,40), ht_Host);

    // Do it
    TPlanet pl = PLANET_DATA;
    runTest("testGrowthMaxTim", pl, EXPECT);
}


void
GamePlanetPredTestSuite::testGrowthTholianTim()
{
    static const int32_t EXPECT[101] = {
            0,  1000,  2000,  3000,  4000,  5000,  6000,  7000,  8000,  9000, 10000, 10000, 10000, 10000, 10000,
        10075, 10080, 10085, 10090, 10095, 10100, 10105, 10110, 10115, 10120, 10125, 10130, 10135, 10140, 10145,
        10150, 10155, 10160, 10165, 10170, 10175, 10180, 10185, 10190, 10195, 10200, 10205, 10210, 10215, 10220,
        10225, 10230, 10235, 10240, 10245, 10250, 10255, 10260, 10265, 10270, 10275, 10280, 10285, 10290, 10295,
        10300, 10305, 10310, 10315, 10320, 10325, 10330, 10335, 10340, 10345, 10350, 10355, 10360, 10365, 10370,
        10375, 10380, 10385, 10390, 10395, 10400, 10405, 10410, 10415, 10420, 10425, 10430, 10435, 10440, 10445,
        10450, 10455, 10460, 10465, 10470, 10475, 10480, 10485, 10490, 10495, 10500
    };

    // Config
    config.assignDefaults();
    config.ClimateDeathRate.set(0);
    config.ClimateLimitsPopulation.set(1);

    // Host
    host.setHostVersion(MKVERSION(3,22,40), ht_Host);

    // Do it
    TPlanet pl = PLANET_DATA;
    pl.owner = 7;
    pl.colonists = 10000;
    runTest("testGrowthTholianTim", pl, EXPECT);
}

void
GamePlanetPredTestSuite::testGrowthRebelTim()
{
    static const int32_t EXPECT[101] = {
        90000, 90000, 90000, 90000, 90000, 90000, 90000, 90000, 90000, 90000, 90000, 90000, 90000, 90000, 90000,
        90000, 90000, 90000, 90000, 90000, 58882, 61390, 63838, 66223, 68543, 70795, 72977, 75088, 77124, 79085,
        80967, 82770, 84491, 86128, 87681, 89148, 90526, 91815, 93014, 94121, 95135, 96056, 96881, 97318, 97334,
        97346, 97356, 97364, 97370, 97374, 97375, 97374, 97370, 97364, 97356, 97346, 97332, 97318, 96842, 96011,
        95086, 94067, 92955, 91752, 90458, 89075, 87605, 86047, 84405, 82680, 80874, 78987, 77023, 74983, 72869,
        70683, 68427, 66104, 63715, 61264, 58753, 56183, 53558, 50881, 48153, 95000, 95000, 95000, 95000, 95000,
        95000, 95000, 95000, 95000, 95000, 95000, 95000, 95000, 95000, 95000, 95000
    };

    // Config
    config.assignDefaults();
    config.ClimateDeathRate.set(0);
    config.ClimateLimitsPopulation.set(1);

    // Host
    host.setHostVersion(MKVERSION(3,22,40), ht_Host);

    // Do it
    TPlanet pl = PLANET_DATA;
    pl.colonists = 95000;
    pl.owner = 10;
    runTest("testGrowthRebelTim", pl, EXPECT);
}

void
GamePlanetPredTestSuite::testGrowthHumanoidTim()
{
    static const int32_t EXPECT[101] = {
        10000, 10000, 10000, 10038, 10051, 10063, 10076, 10088, 10100, 10112, 10124, 10136, 10148, 10159, 10171,
        10182, 10193, 10204, 10215, 10225, 10236, 10246, 10255, 10265, 10274, 10283, 10292, 10300, 10308, 10316,
        10324, 10331, 10338, 10345, 10351, 10357, 10362, 10367, 10372, 10376, 10381, 10384, 10388, 10390, 10393,
        10395, 10397, 10398, 10399, 10400, 10400, 10400, 10399, 10398, 10397, 10395, 10393, 10390, 10387, 10384,
        10380, 10376, 10372, 10367, 10362, 10356, 10350, 10344, 10338, 10331, 10323, 10316, 10308, 10300, 10291,
        10283, 10274, 10264, 10255, 10245, 10235, 10225, 10214, 10204, 10193, 10182, 10170, 10159, 10147, 10135,
        10124, 10112, 10099, 10087, 10075, 10063, 10050, 10038, 10000, 10000, 10000
    };

    // Config
    config.assignDefaults();
    config.ClimateDeathRate.set(0);
    config.ClimateLimitsPopulation.set(1);

    // Host
    host.setHostVersion(MKVERSION(3,22,40), ht_Host);

    // Do it
    TPlanet pl = PLANET_DATA;
    pl.natives = 10000;
    pl.native_race = 1;
    pl.native_gov = 5;
    pl.native_happy = 100;
    runTestNatives("testGrowthHumanoidTim", pl, EXPECT);
}
